# Copyright 2022 Zurich Instruments AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import logging
import math
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

from engineering_notation import EngNumber
from sortedcontainers import SortedDict

from laboneq.compiler.code_generator.feedback_register_allocator import (
    FeedbackRegisterAllocator,
)
from laboneq.compiler.code_generator.utils import resample_state
from laboneq.compiler.common.awg_info import AWGInfo
from laboneq.compiler.common.awg_sampled_event import (
    AWGEvent,
    AWGEventType,
    AWGSampledEventSequence,
)
from laboneq.compiler.common.device_type import DeviceType
from laboneq.compiler.common.event_type import EventType
from laboneq.core.exceptions import LabOneQException
from laboneq.core.utilities.pulse_sampler import interval_to_samples, length_to_samples

if TYPE_CHECKING:
    from laboneq.compiler.common.signal_obj import SignalObj

_logger = logging.getLogger(__name__)


def analyze_loop_times(
    awg: AWGInfo,
    events: List[Any],
    sampling_rate: float,
    delay: float,
) -> AWGSampledEventSequence:
    loop_events = AWGSampledEventSequence()

    plays_anything = False
    signal_ids = [signal_obj.id for signal_obj in awg.signals]
    for e in events:
        if (
            e["event_type"] in ["PLAY_START", "ACQUIRE_START"]
            and e.get("signal") in signal_ids
        ):
            plays_anything = True
            break

    if plays_anything:
        _logger.debug(
            "Analyzing loop events for awg %d of %s", awg.awg_number, awg.device_id
        )
    else:
        _logger.debug(
            "Skipping analysis of loop events for awg %d of %s because nothing is played",
            awg.awg_number,
            awg.device_id,
        )
        return loop_events

    loop_step_start_events = [
        event for event in events if event["event_type"] == "LOOP_STEP_START"
    ]
    loop_step_end_events = [
        event for event in events if event["event_type"] == "LOOP_STEP_END"
    ]
    loop_iteration_events = [
        event
        for event in events
        if event["event_type"] == "LOOP_ITERATION_END"
        and "compressed" in event
        and event["compressed"]
    ]

    compressed_sections = set(
        [event["section_name"] for event in loop_iteration_events]
    )
    section_repeats = {}
    for event in loop_iteration_events:
        if "num_repeats" in event:
            section_repeats[event["section_name"]] = event["num_repeats"]

    _logger.debug("Found %d loop step start events", len(loop_step_start_events))
    events_already_added = set()
    for event in loop_step_start_events:
        _logger.debug("  loop timing: processing  %s", event)

        event_time_in_samples = length_to_samples(event["time"] + delay, sampling_rate)
        loop_event = AWGEvent(
            type=AWGEventType.LOOP_STEP_START,
            start=event_time_in_samples,
            end=event_time_in_samples,
            params={
                "nesting_level": event["nesting_level"],
                "loop_id": event["section_name"],
            },
        )

        if event["section_name"] not in compressed_sections:
            frozen = loop_event.frozen()
            if frozen not in events_already_added:
                loop_events.add(event_time_in_samples, loop_event)
                events_already_added.add(frozen)
                _logger.debug("Added %s", loop_event)
            else:
                _logger.debug("SKIP adding double %s", loop_event)
        elif event["iteration"] == 0:
            push_event = AWGEvent(
                type=AWGEventType.PUSH_LOOP,
                start=event_time_in_samples,
                end=event_time_in_samples,
                params={
                    "nesting_level": event["nesting_level"],
                    "loop_id": event["section_name"],
                    "num_repeats": section_repeats.get(event["section_name"]),
                },
            )
            loop_events.add(event_time_in_samples, push_event)
            _logger.debug("Added %s", push_event)

    _logger.debug("Found %d loop step end events", len(loop_step_end_events))
    for event in loop_step_end_events:
        event_time_in_samples = length_to_samples(event["time"] + delay, sampling_rate)
        loop_event = AWGEvent(
            type=AWGEventType.LOOP_STEP_END,
            start=event_time_in_samples,
            end=event_time_in_samples,
            params={
                "nesting_level": event["nesting_level"],
                "loop_id": event["section_name"],
            },
        )

        if event["section_name"] not in compressed_sections:
            frozen = loop_event.frozen()
            if frozen not in events_already_added:
                loop_events.add(event_time_in_samples, loop_event)
                events_already_added.add(frozen)
                _logger.debug("Added %s", loop_event)
            else:
                _logger.debug("SKIP adding double %s", loop_event)

    for event in loop_iteration_events:
        event_time_in_samples = length_to_samples(event["time"] + delay, sampling_rate)
        iteration_event = AWGEvent(
            type=AWGEventType.ITERATE,
            start=event_time_in_samples,
            end=event_time_in_samples,
            params={
                "nesting_level": event["nesting_level"],
                "loop_id": event["section_name"],
                "num_repeats": section_repeats.get(event["section_name"]),
            },
        )

        loop_events.add(event_time_in_samples, iteration_event)
        _logger.debug("Added %s", iteration_event)

        if event_time_in_samples % awg.device_type.sample_multiple != 0:
            _logger.warning(
                "Event %s: event_time_in_samples %f at sampling rate %s is not divisible by %d",
                event,
                event_time_in_samples,
                EngNumber(sampling_rate),
                awg.device_type.sample_multiple,
            )

    return loop_events


def analyze_init_times(
    device_id: str, sampling_rate: float, delay: float
) -> AWGSampledEventSequence:
    _logger.debug("Analyzing init events for device %s", device_id)
    init_events = AWGSampledEventSequence()
    delay_samples = length_to_samples(delay, sampling_rate)
    init_events.add(
        delay_samples,
        AWGEvent(
            type=AWGEventType.SEQUENCER_START,
            start=delay_samples,
            end=delay_samples,
            params={
                "device_id": device_id,
            },
        ),
    )
    return init_events


def analyze_precomp_reset_times(
    events: List[Any],
    signals: List[str],
    sampling_rate: float,
    delay: float,
) -> Tuple[AWGSampledEventSequence, List[Dict[str, Any]]]:
    precomp_events = [
        event
        for event in events
        if (
            event["event_type"] == EventType.RESET_PRECOMPENSATION_FILTERS
            and event.get("signal_id", None) in signals
        )
    ]
    precomp_reset_events = AWGSampledEventSequence()
    fake_play_events = []
    for event in precomp_events:
        event_time_in_samples = length_to_samples(event["time"] + delay, sampling_rate)

        sampled_event = AWGEvent(
            type=AWGEventType.RESET_PRECOMPENSATION_FILTERS,
            start=event_time_in_samples,
            params={
                "signal_id": event.get("signal_id"),
            },
        )
        precomp_reset_events.add(event_time_in_samples, sampled_event)

        # To force a playWave at the pre-compensation reset time, we add fake
        # PLAY_START and PLAY_END events to the event list. The interval calculator
        # will then pick these up and create a waveform that is all zeros at the reset.

        # Todo: Remove when HULK-1246 is fixed.

        common = {
            "signal": event.get("signal_id"),
            "section_name": event["section_name"],
            "play_wave_id": "dummy_precomp_reset",
            "amplitude": 1.0,  # non-zero amplitude to avoid being ignored
        }

        fake_play_events.extend(
            [
                {
                    "event_type": EventType.PLAY_START,
                    "time": event["time"],
                    **common,
                },
                {
                    "event_type": EventType.PLAY_END,
                    "time": event["time"] + 32 / sampling_rate,
                    **common,
                },
            ]
        )

    return precomp_reset_events, fake_play_events


def analyze_phase_reset_times(
    events: List[Any],
    device_id: str,
    sampling_rate: float,
    delay: float,
) -> AWGSampledEventSequence:
    reset_phase_events = [
        event
        for event in events
        if event["event_type"]
        in (
            EventType.RESET_HW_OSCILLATOR_PHASE,
            EventType.INITIAL_RESET_HW_OSCILLATOR_PHASE,
        )
        and "device_id" in event
        and event["device_id"] == device_id
    ]
    phase_reset_events = AWGSampledEventSequence()
    for event in reset_phase_events:
        event_time_in_samples = length_to_samples(event["time"] + delay, sampling_rate)
        signature = (
            AWGEventType.RESET_PHASE
            if event["event_type"] == EventType.RESET_HW_OSCILLATOR_PHASE
            else AWGEventType.INITIAL_RESET_PHASE
        )
        init_event = AWGEvent(
            type=signature,
            start=event_time_in_samples,
            end=event_time_in_samples,
            params={
                "device_id": device_id,
            },
        )
        phase_reset_events.add(event_time_in_samples, init_event)
    return phase_reset_events


def analyze_set_oscillator_times(
    events: List,
    signal_obj: SignalObj,
) -> AWGSampledEventSequence:
    signal_id = signal_obj.id
    device_id = signal_obj.device_id
    device_type = signal_obj.device_type
    sampling_rate = signal_obj.sampling_rate
    delay = signal_obj.total_delay
    set_oscillator_events = [
        event
        for event in events
        if event["event_type"] == "SET_OSCILLATOR_FREQUENCY_START"
        and event.get("device_id") == device_id
        and event.get("signal") == signal_id
    ]
    if len(set_oscillator_events) == 0:
        return AWGSampledEventSequence()

    if device_type not in (DeviceType.HDAWG, DeviceType.SHFQA, DeviceType.SHFSG):
        raise LabOneQException(
            "Real-time frequency sweep only supported on SHF and HDAWG devices"
        )

    iterations = {event["iteration"]: event for event in set_oscillator_events}
    assert list(iterations.keys()) == list(
        range(len(iterations))
    )  # "iteration" values are unique, ordered, and numbered 0 .. N-1
    start_frequency = iterations[0]["value"]
    if len(iterations) > 1:
        step_frequency = iterations[1]["value"] - start_frequency
    else:
        step_frequency = 0

    retval = AWGSampledEventSequence()

    for iteration, event in iterations.items():
        if (
            abs(event["value"] - iteration * step_frequency - start_frequency)
            > 1e-3  # tolerance: 1 mHz
        ):
            raise LabOneQException("Realtime oscillator sweeps must be linear")

        event_time_in_samples = length_to_samples(event["time"] + delay, sampling_rate)
        set_oscillator_event = AWGEvent(
            type=AWGEventType.SET_OSCILLATOR_FREQUENCY,
            start=event_time_in_samples,
            params={
                "start_frequency": start_frequency,
                "step_frequency": step_frequency,
                "parameter_name": event["parameter"]["id"],
                "iteration": iteration,
                "iterations": len(iterations),
            },
        )

        retval.add(event_time_in_samples, set_oscillator_event)

    return retval


def analyze_acquire_times(
    events: List[Any],
    signal_obj: SignalObj,
    feedback_register_allocator: FeedbackRegisterAllocator,
) -> AWGSampledEventSequence:

    signal_id = signal_obj.id
    sampling_rate = signal_obj.sampling_rate
    delay = signal_obj.total_delay
    sample_multiple = signal_obj.device_type.sample_multiple
    channels = signal_obj.channels

    _logger.debug(
        "Calculating acquire times for signal %s with delay %s ( %s samples)",
        signal_id,
        str(delay),
        str(round(delay * sampling_rate)),
    )

    @dataclass
    class IntervalStartEvent:
        event_type: str
        time: float
        play_wave_id: str
        acquisition_type: list
        acquire_handle: str
        play_pulse_parameters: Optional[Dict[str, Any]]
        pulse_pulse_parameters: Optional[Dict[str, Any]]

    @dataclass
    class IntervalEndEvent:
        event_type: str
        time: float
        play_wave_id: str

    interval_zip = list(
        zip(
            [
                IntervalStartEvent(
                    event["event_type"],
                    event["time"] + delay,
                    event["play_wave_id"],
                    event.get("acquisition_type", []),
                    event["acquire_handle"],
                    event.get("play_pulse_parameters"),
                    event.get("pulse_pulse_parameters"),
                )
                for event in events
                if event["event_type"] in ["ACQUIRE_START"]
                and event["signal"] == signal_id
            ],
            [
                IntervalEndEvent(
                    event["event_type"],
                    event["time"] + delay,
                    event["play_wave_id"],
                )
                for event in events
                if event["event_type"] in ["ACQUIRE_END"]
                and event["signal"] == signal_id
            ],
        )
    )

    acquire_events = AWGSampledEventSequence()
    for interval_start, interval_end in interval_zip:
        start_samples, end_samples = interval_to_samples(
            interval_start.time, interval_end.time, sampling_rate
        )
        if start_samples % sample_multiple != 0:
            start_samples = (
                math.floor(start_samples / sample_multiple) * sample_multiple
            )

        if end_samples % sample_multiple != 0:
            end_samples = math.floor(end_samples / sample_multiple) * sample_multiple

        feedback_register = (
            None
            if feedback_register_allocator is None
            else feedback_register_allocator.allocate(
                signal_id, interval_start.acquire_handle
            )
        )

        acquire_event = AWGEvent(
            type=AWGEventType.ACQUIRE,
            start=start_samples,
            end=end_samples,
            params={
                "signal_id": signal_id,
                "play_wave_id": interval_start.play_wave_id,
                "acquisition_type": interval_start.acquisition_type,
                "acquire_handles": [interval_start.acquire_handle],
                "feedback_register": feedback_register,
                "channels": channels,
                "play_pulse_parameters": interval_start.play_pulse_parameters,
                "pulse_pulse_parameters": interval_start.pulse_pulse_parameters,
            },
        )

        acquire_events.add(start_samples, acquire_event)

    return acquire_events


def analyze_trigger_events(
    events: List[Dict], signal: SignalObj, loop_events: AWGSampledEventSequence
) -> AWGSampledEventSequence:
    digital_signal_change_events = [
        event
        for event in events
        if event["event_type"] == EventType.DIGITAL_SIGNAL_STATE_CHANGE
        and signal.id == event["signal"]
    ]
    delay = signal.total_delay
    sampling_rate = signal.sampling_rate
    device_type = signal.awg.device_type

    sampled_digital_signal_change_events = AWGSampledEventSequence()

    for event in digital_signal_change_events:
        time_in_samples = length_to_samples(event["time"] + delay, sampling_rate)
        sampled_digital_signal_change_events.add(
            time_in_samples, AWGEvent(type=None, params=event)
        )

    current_state = 0
    retval = AWGSampledEventSequence()
    state_progression = SortedDict()

    event_list: List[AWGEvent]
    for (
        time_in_samples,
        event_list,
    ) in sampled_digital_signal_change_events.sequence.items():
        if device_type in (DeviceType.SHFQA, DeviceType.SHFSG):
            for event in event_list:
                if event.params["bit"] > 0:
                    raise LabOneQException(
                        f"On device {device_type.value}, only a single trigger channel is "
                        f"available (section {event.params['section_name']})."
                    )
        for event in [e for e in event_list if e.params["change"] == "CLEAR"]:
            mask = ~(2 ** event.params["bit"])
            current_state = current_state & mask
        for event in [e for e in event_list if e.params["change"] == "SET"]:
            mask = 2 ** event.params["bit"]
            current_state = current_state | mask
        state_progression[time_in_samples] = current_state
        retval.add(
            time_in_samples,
            AWGEvent(
                type=AWGEventType.TRIGGER_OUTPUT,
                start=time_in_samples,
                params={
                    "state": current_state,
                },
            ),
        )

    if state_progression and 0 not in state_progression:
        # assume that at time 0, the state is 0
        state_progression[0] = 0

    # When the trigger is raised at the end of the averaging loop (which is not
    # unrolled), things get a bit dicey: the command to reset the trigger signal must
    # be deferred to the next iteration. Which means that the very first iteration must
    # already include this tigger reset, and that we must issue it again after the loop.

    resampled_states = resample_state(loop_events.sequence.keys(), state_progression)
    if resampled_states:
        for time_in_samples, event_list in loop_events.sequence.items():
            if any(event.type == AWGEventType.PUSH_LOOP for event in event_list):
                retval.add(
                    time_in_samples,
                    AWGEvent(
                        type=AWGEventType.TRIGGER_OUTPUT,
                        start=time_in_samples,
                        params={
                            "state": resampled_states[time_in_samples][1],
                        },
                    ),
                )
    return retval
